/*
 * Copyright 2010 Linaro
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation.
 *
 * Authored by: Jammy Zhou <jammy.zhou@linaro.org>
 *
 */
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <dlfcn.h>
#include <GL/gl_proxy.h>
#include <GL/glx_proxy.h>

/* global proxy context */
PROXY_CONTEXT proxy_context;

/* library names for different backends */
const char *opengl_libraries[] = {
	"/usr/lib/libGL.so",
	"/usr/lib/libGLESv2.so",
};

/* proxy functions */
PFNPROXYACTIVETEXTUREPROC pfnProxyActiveTexture = NULL;
PFNPROXYATTACHSHADERPROC pfnProxyAttachShader = NULL;
PFNPROXYBINDATTRIBLOCATIONPROC pfnProxyBindAttribLocation = NULL;
PFNPROXYBINDBUFFERPROC pfnProxyBindBuffer = NULL;
PFNPROXYBINDFRAMEBUFFERPROC pfnProxyBindFramebuffer = NULL;
PFNPROXYBINDRENDERBUFFERPROC pfnProxyBindRenderbuffer = NULL;
PFNPROXYBINDTEXTUREPROC pfnProxyBindTexture = NULL;
PFNPROXYBLENDCOLORPROC pfnProxyBlendColor = NULL;
PFNPROXYBLENDEQUATIONPROC pfnProxyBlendEquation = NULL;
PFNPROXYBLENDEQUATIONSEPARATEPROC pfnProxyBlendEquationSeparate = NULL;
PFNPROXYBLENDFUNCPROC pfnProxyBlendFunc = NULL;
PFNPROXYBLENDFUNCSEPARATEPROC pfnProxyBlendFuncSeparate = NULL;
PFNPROXYBUFFERDATAPROC pfnProxyBufferData = NULL;
PFNPROXYBUFFERSUBDATAPROC pfnProxyBufferSubData = NULL;
PFNPROXYCHECKFRAMEBUFFERSTATUSPROC pfnProxyCheckFramebufferStatus = NULL;
PFNPROXYCLEARPROC pfnProxyClear = NULL;
PFNPROXYCLEARCOLORPROC pfnProxyClearColor = NULL;
PFNPROXYCLEARDEPTHFPROC pfnProxyClearDepthf = NULL;
PFNPROXYCLEARSTENCILPROC pfnProxyClearStencil = NULL;
PFNPROXYCOLORMASKPROC pfnProxyColorMask = NULL;
PFNPROXYCOMPILESHADERPROC pfnProxyCompileShader = NULL;
PFNPROXYCOMPRESSEDTEXIMAGE2DPROC pfnProxyCompressedTexImage2D = NULL;
PFNPROXYCOMPRESSEDTEXSUBIMAGE2DPROC pfnProxyCompressedTexSubImage2D = NULL;
PFNPROXYCOPYTEXIMAGE2DPROC pfnProxyCopyTexImage2D = NULL;
PFNPROXYCOPYTEXSUBIMAGE2DPROC pfnProxyCopyTexSubImage2D = NULL;
PFNPROXYCREATEPROGRAMPROC pfnProxyCreateProgram = NULL;
PFNPROXYCREATESHADERPROC pfnProxyCreateShader = NULL;
PFNPROXYCULLFACEPROC pfnProxyCullFace = NULL;
PFNPROXYDELETEBUFFERSPROC pfnProxyDeleteBuffers = NULL;
PFNPROXYDELETEFRAMEBUFFERSPROC pfnProxyDeleteFramebuffers = NULL;
PFNPROXYDELETEPROGRAMPROC pfnProxyDeleteProgram = NULL;
PFNPROXYDELETERENDERBUFFERSPROC pfnProxyDeleteRenderbuffers = NULL;
PFNPROXYDELETESHADERPROC pfnProxyDeleteShader = NULL;
PFNPROXYDELETETEXTURESPROC pfnProxyDeleteTextures = NULL;
PFNPROXYDEPTHFUNCPROC pfnProxyDepthFunc = NULL;
PFNPROXYDEPTHMASKPROC pfnProxyDepthMask = NULL;
PFNPROXYDEPTHRANGEFPROC pfnProxyDepthRangef = NULL;
PFNPROXYDETACHSHADERPROC pfnProxyDetachShader = NULL;
PFNPROXYDISABLEPROC pfnProxyDisable = NULL;
PFNPROXYDISABLEVERTEXATTRIBARRAYPROC pfnProxyDisableVertexAttribArray = NULL;
PFNPROXYDRAWARRAYSPROC pfnProxyDrawArrays = NULL;
PFNPROXYDRAWELEMENTSPROC pfnProxyDrawElements = NULL;
PFNPROXYENABLEPROC pfnProxyEnable = NULL;
PFNPROXYENABLEVERTEXATTRIBARRAYPROC pfnProxyEnableVertexAttribArray = NULL;
PFNPROXYFINISHPROC pfnProxyFinish = NULL;
PFNPROXYFLUSHPROC pfnProxyFlush = NULL;
PFNPROXYFRAMEBUFFERRENDERBUFFERPROC pfnProxyFramebufferRenderbuffer = NULL;
PFNPROXYFRAMEBUFFERTEXTURE2DPROC pfnProxyFramebufferTexture2D = NULL;
PFNPROXYFRONTFACEPROC pfnProxyFrontFace = NULL;
PFNPROXYGENBUFFERSPROC pfnProxyGenBuffers = NULL;
PFNPROXYGENERATEMIPMAPPROC pfnProxyGenerateMipmap = NULL;
PFNPROXYGENFRAMEBUFFERSPROC pfnProxyGenFramebuffers = NULL;
PFNPROXYGENRENDERBUFFERSPROC pfnProxyGenRenderbuffers = NULL;
PFNPROXYGENTEXTURESPROC pfnProxyGenTextures = NULL;
PFNPROXYGETACTIVEATTRIBPROC pfnProxyGetActiveAttrib = NULL;
PFNPROXYGETACTIVEUNIFORMPROC pfnProxyGetActiveUniform = NULL;
PFNPROXYGETATTACHEDSHADERSPROC pfnProxyGetAttachedShaders = NULL;
PFNPROXYGETATTRIBLOCATIONPROC pfnProxyGetAttribLocation = NULL;
PFNPROXYGETBOOLEANVPROC pfnProxyGetBooleanv = NULL;
PFNPROXYGETBUFFERPARAMETERIVPROC pfnProxyGetBufferParameteriv = NULL;
PFNPROXYGETERRORPROC pfnProxyGetError = NULL;
PFNPROXYGETFLOATVPROC pfnProxyGetFloatv = NULL;
PFNPROXYGETFRAMEBUFFERATTACHMENTPARAMETERIVPROC pfnProxyGetFramebufferAttachmentParameteriv = NULL;
PFNPROXYGETINTEGERVPROC pfnProxyGetIntegerv = NULL;
PFNPROXYGETPROGRAMIVPROC pfnProxyGetProgramiv = NULL;
PFNPROXYGETPROGRAMINFOLOGPROC pfnProxyGetProgramInfoLog = NULL;
PFNPROXYGETRENDERBUFFERPARAMETERIVPROC pfnProxyGetRenderbufferParameteriv = NULL;
PFNPROXYGETSHADERIVPROC pfnProxyGetShaderiv = NULL;
PFNPROXYGETSHADERINFOLOGPROC pfnProxyGetShaderInfoLog = NULL;
PFNPROXYGETSHADERPRECISIONFORMATPROC pfnProxyGetShaderPrecisionFormat = NULL;
PFNPROXYGETSHADERSOURCEPROC pfnProxyGetShaderSource = NULL;
PFNPROXYGETSTRINGPROC pfnProxyGetString = NULL;
PFNPROXYGETTEXPARAMETERFVPROC pfnProxyGetTexParameterfv = NULL;
PFNPROXYGETTEXPARAMETERIVPROC pfnProxyGetTexParameteriv = NULL;
PFNPROXYGETUNIFORMFVPROC pfnProxyGetUniformfv = NULL;
PFNPROXYGETUNIFORMIVPROC pfnProxyGetUniformiv = NULL;
PFNPROXYGETUNIFORMLOCATIONPROC pfnProxyGetUniformLocation = NULL;
PFNPROXYGETVERTEXATTRIBFVPROC pfnProxyGetVertexAttribfv = NULL;
PFNPROXYGETVERTEXATTRIBIVPROC pfnProxyGetVertexAttribiv = NULL;
PFNPROXYGETVERTEXATTRIBPOINTERVPROC pfnProxyGetVertexAttribPointerv = NULL;
PFNPROXYHINTPROC pfnProxyHint = NULL;
PFNPROXYISBUFFERPROC pfnProxyIsBuffer = NULL;
PFNPROXYISENABLEDPROC pfnProxyIsEnabled = NULL;
PFNPROXYISFRAMEBUFFERPROC pfnProxyIsFramebuffer = NULL;
PFNPROXYISPROGRAMPROC pfnProxyIsProgram = NULL;
PFNPROXYISRENDERBUFFERPROC pfnProxyIsRenderbuffer = NULL;
PFNPROXYISSHADERPROC pfnProxyIsShader = NULL;
PFNPROXYISTEXTUREPROC pfnProxyIsTexture = NULL;
PFNPROXYLINEWIDTHPROC pfnProxyLineWidth = NULL;
PFNPROXYLINKPROGRAMPROC pfnProxyLinkProgram = NULL;
PFNPROXYPIXELSTOREIPROC pfnProxyPixelStorei = NULL;
PFNPROXYPOLYGONOFFSETPROC pfnProxyPolygonOffset = NULL;
PFNPROXYREADPIXELSPROC pfnProxyReadPixels = NULL;
PFNPROXYRELEASESHADERCOMPILERPROC pfnProxyReleaseShaderCompiler = NULL;
PFNPROXYRENDERBUFFERSTORAGEPROC pfnProxyRenderbufferStorage = NULL;
PFNPROXYSAMPLECOVERAGEPROC pfnProxySampleCoverage = NULL;
PFNPROXYSCISSORPROC pfnProxyScissor = NULL;
PFNPROXYSHADERBINARYPROC pfnProxyShaderBinary = NULL;
PFNPROXYSHADERSOURCEPROC pfnProxyShaderSource = NULL;
PFNPROXYSTENCILFUNCPROC pfnProxyStencilFunc = NULL;
PFNPROXYSTENCILFUNCSEPARATEPROC pfnProxyStencilFuncSeparate = NULL;
PFNPROXYSTENCILMASKPROC pfnProxyStencilMask = NULL;
PFNPROXYSTENCILMASKSEPARATEPROC pfnProxyStencilMaskSeparate = NULL;
PFNPROXYSTENCILOPPROC pfnProxyStencilOp = NULL;
PFNPROXYSTENCILOPSEPARATEPROC pfnProxyStencilOpSeparate = NULL;
PFNPROXYTEXIMAGE2DPROC pfnProxyTexImage2D = NULL;
PFNPROXYTEXPARAMETERFPROC pfnProxyTexParameterf = NULL;
PFNPROXYTEXPARAMETERFVPROC pfnProxyTexParameterfv = NULL;
PFNPROXYTEXPARAMETERIPROC pfnProxyTexParameteri = NULL;
PFNPROXYTEXPARAMETERIVPROC pfnProxyTexParameteriv = NULL;
PFNPROXYTEXSUBIMAGE2DPROC pfnProxyTexSubImage2D = NULL;
PFNPROXYUNIFORM1FPROC pfnProxyUniform1f = NULL;
PFNPROXYUNIFORM1FVPROC pfnProxyUniform1fv = NULL;
PFNPROXYUNIFORM1IPROC pfnProxyUniform1i = NULL;
PFNPROXYUNIFORM1IVPROC pfnProxyUniform1iv = NULL;
PFNPROXYUNIFORM2FPROC pfnProxyUniform2f = NULL;
PFNPROXYUNIFORM2FVPROC pfnProxyUniform2fv = NULL;
PFNPROXYUNIFORM2IPROC pfnProxyUniform2i = NULL;
PFNPROXYUNIFORM2IVPROC pfnProxyUniform2iv = NULL;
PFNPROXYUNIFORM3FPROC pfnProxyUniform3f = NULL;
PFNPROXYUNIFORM3FVPROC pfnProxyUniform3fv = NULL;
PFNPROXYUNIFORM3IPROC pfnProxyUniform3i = NULL;
PFNPROXYUNIFORM3IVPROC pfnProxyUniform3iv = NULL;
PFNPROXYUNIFORM4FPROC pfnProxyUniform4f = NULL;
PFNPROXYUNIFORM4FVPROC pfnProxyUniform4fv = NULL;
PFNPROXYUNIFORM4IPROC pfnProxyUniform4i = NULL;
PFNPROXYUNIFORM4IVPROC pfnProxyUniform4iv = NULL;
PFNPROXYUNIFORMMATRIX2FVPROC pfnProxyUniformMatrix2fv = NULL;
PFNPROXYUNIFORMMATRIX3FVPROC pfnProxyUniformMatrix3fv = NULL;
PFNPROXYUNIFORMMATRIX4FVPROC pfnProxyUniformMatrix4fv = NULL;
PFNPROXYUSEPROGRAMPROC pfnProxyUseProgram = NULL;
PFNPROXYVALIDATEPROGRAMPROC pfnProxyValidateProgram = NULL;
PFNPROXYVERTEXATTRIB1FPROC pfnProxyVertexAttrib1f = NULL;
PFNPROXYVERTEXATTRIB1FVPROC pfnProxyVertexAttrib1fv = NULL;
PFNPROXYVERTEXATTRIB2FPROC pfnProxyVertexAttrib2f = NULL;
PFNPROXYVERTEXATTRIB2FVPROC pfnProxyVertexAttrib2fv = NULL;
PFNPROXYVERTEXATTRIB3FPROC pfnProxyVertexAttrib3f = NULL;
PFNPROXYVERTEXATTRIB3FVPROC pfnProxyVertexAttrib3fv = NULL;
PFNPROXYVERTEXATTRIB4FPROC pfnProxyVertexAttrib4f = NULL;
PFNPROXYVERTEXATTRIB4FVPROC pfnProxyVertexAttrib4fv = NULL;
PFNPROXYVERTEXATTRIBPOINTERPROC pfnProxyVertexAttribPointer = NULL;
PFNPROXYVIEWPORTPROC pfnProxyViewport = NULL;

/* Proc address for EGL/GLX symbol resolve function
 * Note:
 * eglGetProcAddress can only be used for GL or EGL extension function
 * glXGetProcAddress/glXGetProcAddressARB can be used for all GL/GLX funtions
 */
typedef void * (*PFNPROXYEGLGETPROCADDRESSPROC)( char const *procName);

PFNPROXYGLXGETPROCADDRESSPROC pfnProxyGLXGetProcAddressARB = NULL;
PFNPROXYGLXGETPROCADDRESSPROC pfnProxyGLXGetProcAddress = NULL;
PFNPROXYEGLGETPROCADDRESSPROC pfnProxyEGLGetProcAddress = NULL;

static void *glProxyOpenLibrary(const char *name)
{
	return dlopen(name, RTLD_LAZY | RTLD_LOCAL);
}

static int glProxyCloseLibrary(void *handle)
{
	return dlclose(handle);
}

static void glProxyEarlyResolve(void)
{
	if (OPENGL_BACKEND == proxy_context.current_backend) {
		pfnProxyGLXGetProcAddress = dlsym(proxy_context.gl_handle, "glXGetProcAddress");
		if (NULL == pfnProxyGLXGetProcAddress) {
			pfnProxyGLXGetProcAddressARB = dlsym(proxy_context.gl_handle, "glXGetProcAddressARB");
		}
	}
	else if (OPENGL_ES20_BACKEND == proxy_context.current_backend) {
		pfnProxyEGLGetProcAddress = dlsym(proxy_context.egl_handle, "eglGetProcAddress");
	}
}

void *glProxyGetProcAddr(void *handle, const char *name)
{
	if (pfnProxyGLXGetProcAddress) {
		return (*pfnProxyGLXGetProcAddress)(name);
	}
	else if (pfnProxyGLXGetProcAddressARB) {
		return (*pfnProxyGLXGetProcAddressARB)(name);
	}
	else {
		return dlsym(handle, name);
	}
}

void *glProxyGetExtProcAddr(void *handle, const char *name)
{
	if (pfnProxyEGLGetProcAddress) {
		return (*pfnProxyEGLGetProcAddress)(name);
	}
	else if (pfnProxyGLXGetProcAddress) {
		return (*pfnProxyGLXGetProcAddress)(name);
	}
	else if (pfnProxyGLXGetProcAddressARB) {
		return (*pfnProxyGLXGetProcAddressARB)(name);
	}
	else {
		return dlsym(handle, name);
	}
}

static void *glProxyBackendSelection(PROXY_BACKEND_OPTION *pProxyBackend)
{
	PROXY_BACKEND_OPTION option = *pProxyBackend;
	void *handle = NULL;

	if (OPENGL_BACKEND == option) {
		handle = glProxyOpenLibrary(opengl_libraries[OPENGL_BACKEND]);
		if (NULL == handle) {
			handle = glProxyOpenLibrary(opengl_libraries[OPENGL_ES20_BACKEND]);
			if (NULL == handle) {
				*pProxyBackend = UNKNOWN_BACKEND;
				return NULL;
			} else {
				*pProxyBackend = OPENGL_ES20_BACKEND;
				return handle;
			}
		} else {
			*pProxyBackend = OPENGL_BACKEND;
			return handle;
		}
	} else if (OPENGL_ES20_BACKEND == option) {
		handle = glProxyOpenLibrary(opengl_libraries[OPENGL_ES20_BACKEND]);
                if (NULL == handle) {
                        handle = glProxyOpenLibrary(opengl_libraries[OPENGL_BACKEND]);
                        if (NULL == handle) {
                                *pProxyBackend = UNKNOWN_BACKEND;
                                return NULL;
                        } else {
                                *pProxyBackend = OPENGL_BACKEND;
                                return handle;
                        }
                } else {
                        *pProxyBackend = OPENGL_ES20_BACKEND;
                        return handle;
                }
	} else {
		void *opengl_handle = glProxyOpenLibrary(opengl_libraries[OPENGL_BACKEND]);
		void *gles2_handle = glProxyOpenLibrary(opengl_libraries[OPENGL_ES20_BACKEND]);

		if ((NULL != opengl_handle) && (NULL != gles2_handle)) {
			/*
			 * Because glGetString(GL_RENDERER) only works after rendering context
			 * is created, hw/sw impl. will not be checked here to avoid burden
			 * the implementation, and GLES2 backend will be used in this case.
			 * If OpenGL backend needed, application should specify explicitly.
			 */
			*pProxyBackend = OPENGL_ES20_BACKEND;
			glProxyCloseLibrary(opengl_handle);
			return gles2_handle;
		} else if (NULL != opengl_handle) {
			*pProxyBackend = OPENGL_BACKEND;
			return opengl_handle;
		} else if (NULL != gles2_handle) {
			*pProxyBackend = OPENGL_ES20_BACKEND;
			return gles2_handle;
		} else {
			*pProxyBackend = UNKNOWN_BACKEND;
			return NULL;
		}
	}
}

extern int glxProxyInit(void *handle);

static int glProxyCoreInit(void *handle)
{
	if (NULL == handle) {
		return -1;
	}

	pfnProxyActiveTexture = glProxyGetProcAddr(handle, "glActiveTexture");
	pfnProxyAttachShader = glProxyGetProcAddr(handle, "glAttachShader");
	pfnProxyBindAttribLocation = glProxyGetProcAddr(handle, "glBindAttribLocation");
	pfnProxyBindBuffer = glProxyGetProcAddr(handle, "glBindBuffer");
	pfnProxyBindFramebuffer = glProxyGetProcAddr(handle, "glBindFramebuffer");
	pfnProxyBindRenderbuffer = glProxyGetProcAddr(handle, "glBindRenderbuffer");
	pfnProxyBindTexture = glProxyGetProcAddr(handle, "glBindTexture");
	pfnProxyBlendColor = glProxyGetProcAddr(handle, "glBlendColor");
	pfnProxyBlendEquation = glProxyGetProcAddr(handle, "glBlendEquation");
	pfnProxyBlendEquationSeparate = glProxyGetProcAddr(handle, "glBlendEquationSeparate");
	pfnProxyBlendFunc = glProxyGetProcAddr(handle, "glBlendFunc");
	pfnProxyBlendFuncSeparate = glProxyGetProcAddr(handle, "glBlendFuncSeparate");
	pfnProxyBufferData = glProxyGetProcAddr(handle, "glBufferData");
	pfnProxyBufferSubData = glProxyGetProcAddr(handle, "glBufferSubData");
	pfnProxyCheckFramebufferStatus = glProxyGetProcAddr(handle, "glCheckFramebufferStatus");
	pfnProxyClear = glProxyGetProcAddr(handle, "glClear");
	pfnProxyClearColor = glProxyGetProcAddr(handle, "glClearColor");
	pfnProxyClearDepthf = glProxyGetProcAddr(handle, "glClearDepthf");
	pfnProxyClearStencil = glProxyGetProcAddr(handle, "glClearStencil");
	pfnProxyColorMask = glProxyGetProcAddr(handle, "glColorMask");
	pfnProxyCompileShader = glProxyGetProcAddr(handle, "glCompileShader");
	pfnProxyCompressedTexImage2D = glProxyGetProcAddr(handle, "glCompressedTexImage2D");
	pfnProxyCompressedTexSubImage2D = glProxyGetProcAddr(handle, "glCompressedTexSubImage2D");
	pfnProxyCopyTexImage2D = glProxyGetProcAddr(handle, "glCopyTexImage2D");
	pfnProxyCopyTexSubImage2D = glProxyGetProcAddr(handle, "glCopyTexSubImage2D");
	pfnProxyCreateProgram = glProxyGetProcAddr(handle, "glCreateProgram");
	pfnProxyCreateShader = glProxyGetProcAddr(handle, "glCreateShader");
	pfnProxyCullFace = glProxyGetProcAddr(handle, "glCullFace");
	pfnProxyDeleteBuffers = glProxyGetProcAddr(handle, "glDeleteBuffers");
	pfnProxyDeleteFramebuffers = glProxyGetProcAddr(handle, "glDeleteFramebuffers");
	pfnProxyDeleteProgram = glProxyGetProcAddr(handle, "glDeleteProgram");
	pfnProxyDeleteRenderbuffers = glProxyGetProcAddr(handle, "glDeleteRenderbuffers");
	pfnProxyDeleteShader = glProxyGetProcAddr(handle, "glDeleteShader");
	pfnProxyDeleteTextures = glProxyGetProcAddr(handle, "glDeleteTextures");
	pfnProxyDepthFunc = glProxyGetProcAddr(handle, "glDepthFunc");
	pfnProxyDepthMask = glProxyGetProcAddr(handle, "glDepthMask");
	pfnProxyDepthRangef = glProxyGetProcAddr(handle, "glDepthRangef");
	pfnProxyDetachShader = glProxyGetProcAddr(handle, "glDetachShader");
	pfnProxyDisable = glProxyGetProcAddr(handle, "glDisable");
	pfnProxyDisableVertexAttribArray = glProxyGetProcAddr(handle, "glDisableVertexAttribArray");
	pfnProxyDrawArrays = glProxyGetProcAddr(handle, "glDrawArrays");
	pfnProxyDrawElements = glProxyGetProcAddr(handle, "glDrawElements");
	pfnProxyEnable = glProxyGetProcAddr(handle, "glEnable");
	pfnProxyEnableVertexAttribArray = glProxyGetProcAddr(handle, "glEnableVertexAttribArray");
	pfnProxyFinish = glProxyGetProcAddr(handle, "glFinish");
	pfnProxyFlush = glProxyGetProcAddr(handle, "glFlush");
	pfnProxyFramebufferRenderbuffer = glProxyGetProcAddr(handle, "glFramebufferRenderbuffer");
	pfnProxyFramebufferTexture2D = glProxyGetProcAddr(handle, "glFramebufferTexture2D");
	pfnProxyFrontFace = glProxyGetProcAddr(handle, "glFrontFace");
	pfnProxyGenBuffers = glProxyGetProcAddr(handle, "glGenBuffers");
	pfnProxyGenerateMipmap = glProxyGetProcAddr(handle, "glGenerateMipmap");
	pfnProxyGenFramebuffers = glProxyGetProcAddr(handle, "glGenFramebuffers");
	pfnProxyGenRenderbuffers = glProxyGetProcAddr(handle, "glGenRenderbuffers");
	pfnProxyGenTextures = glProxyGetProcAddr(handle, "glGenTextures");
	pfnProxyGetActiveAttrib = glProxyGetProcAddr(handle, "glGetActiveAttrib");
	pfnProxyGetActiveUniform = glProxyGetProcAddr(handle, "glGetActiveUniform");
	pfnProxyGetAttachedShaders = glProxyGetProcAddr(handle, "glGetAttachedShaders");
	pfnProxyGetAttribLocation = glProxyGetProcAddr(handle, "glGetAttribLocation");
	pfnProxyGetBooleanv = glProxyGetProcAddr(handle, "glGetBooleanv");
	pfnProxyGetBufferParameteriv = glProxyGetProcAddr(handle, "glGetBufferParameteriv");
	pfnProxyGetError = glProxyGetProcAddr(handle, "glGetError");
	pfnProxyGetFloatv = glProxyGetProcAddr(handle, "glGetFloatv");
	pfnProxyGetFramebufferAttachmentParameteriv = glProxyGetProcAddr(handle, "glGetFramebufferAttachmentParameteriv");
	pfnProxyGetIntegerv = glProxyGetProcAddr(handle, "glGetIntegerv");
	pfnProxyGetProgramiv = glProxyGetProcAddr(handle, "glGetProgramiv");
	pfnProxyGetProgramInfoLog = glProxyGetProcAddr(handle, "glGetProgramInfoLog");
	pfnProxyGetRenderbufferParameteriv = glProxyGetProcAddr(handle, "glGetRenderbufferParameteriv");
	pfnProxyGetShaderiv = glProxyGetProcAddr(handle, "glGetShaderiv");
	pfnProxyGetShaderInfoLog = glProxyGetProcAddr(handle, "glGetShaderInfoLog");
	pfnProxyGetShaderPrecisionFormat = glProxyGetProcAddr(handle, "glGetShaderPrecisionFormat");
	pfnProxyGetShaderSource = glProxyGetProcAddr(handle, "glGetShaderSource");
	pfnProxyGetString = glProxyGetProcAddr(handle, "glGetString");
	pfnProxyGetTexParameterfv = glProxyGetProcAddr(handle, "glGetTexParameterfv");
	pfnProxyGetTexParameteriv = glProxyGetProcAddr(handle, "glGetTexParameteriv");
	pfnProxyGetUniformfv = glProxyGetProcAddr(handle, "glGetUniformfv");
	pfnProxyGetUniformiv = glProxyGetProcAddr(handle, "glGetUniformiv");
	pfnProxyGetUniformLocation = glProxyGetProcAddr(handle, "glGetUniformLocation");
	pfnProxyGetVertexAttribfv = glProxyGetProcAddr(handle, "glGetVertexAttribfv");
	pfnProxyGetVertexAttribiv = glProxyGetProcAddr(handle, "glGetVertexAttribiv");
	pfnProxyGetVertexAttribPointerv = glProxyGetProcAddr(handle, "glGetVertexAttribPointerv");
	pfnProxyHint = glProxyGetProcAddr(handle, "glHint");
	pfnProxyIsBuffer = glProxyGetProcAddr(handle, "glIsBuffer");
	pfnProxyIsEnabled = glProxyGetProcAddr(handle, "glIsEnabled");
	pfnProxyIsFramebuffer = glProxyGetProcAddr(handle, "glIsFramebuffer");
	pfnProxyIsProgram = glProxyGetProcAddr(handle, "glIsProgram");
	pfnProxyIsRenderbuffer = glProxyGetProcAddr(handle, "glIsRenderbuffer");
	pfnProxyIsShader = glProxyGetProcAddr(handle, "glIsShader");
	pfnProxyIsTexture = glProxyGetProcAddr(handle, "glIsTexture");
	pfnProxyLineWidth = glProxyGetProcAddr(handle, "glLineWidth");
	pfnProxyLinkProgram = glProxyGetProcAddr(handle, "glLinkProgram");
	pfnProxyPixelStorei = glProxyGetProcAddr(handle, "glPixelStorei");
	pfnProxyPolygonOffset = glProxyGetProcAddr(handle, "glPolygonOffset");
	pfnProxyReadPixels = glProxyGetProcAddr(handle, "glReadPixels");
	pfnProxyReleaseShaderCompiler = glProxyGetProcAddr(handle, "glReleaseShaderCompiler");
	pfnProxyRenderbufferStorage = glProxyGetProcAddr(handle, "glRenderbufferStorage");
	pfnProxySampleCoverage = glProxyGetProcAddr(handle, "glSampleCoverage");
	pfnProxyScissor = glProxyGetProcAddr(handle, "glScissor");
	pfnProxyShaderBinary = glProxyGetProcAddr(handle, "glShaderBinary");
	pfnProxyShaderSource = glProxyGetProcAddr(handle, "glShaderSource");
	pfnProxyStencilFunc = glProxyGetProcAddr(handle, "glStencilFunc");
	pfnProxyStencilFuncSeparate = glProxyGetProcAddr(handle, "glStencilFuncSeparate");
	pfnProxyStencilMask = glProxyGetProcAddr(handle, "glStencilMask");
	pfnProxyStencilMaskSeparate = glProxyGetProcAddr(handle, "glStencilMaskSeparate");
	pfnProxyStencilOp = glProxyGetProcAddr(handle, "glStencilOp");
	pfnProxyStencilOpSeparate = glProxyGetProcAddr(handle, "glStencilOpSeparate");
	pfnProxyTexImage2D = glProxyGetProcAddr(handle, "glTexImage2D");
	pfnProxyTexParameterf = glProxyGetProcAddr(handle, "glTexParameterf");
	pfnProxyTexParameterfv = glProxyGetProcAddr(handle, "glTexParameterfv");
	pfnProxyTexParameteri = glProxyGetProcAddr(handle, "glTexParameteri");
	pfnProxyTexParameteriv = glProxyGetProcAddr(handle, "glTexParameteriv");
	pfnProxyTexSubImage2D = glProxyGetProcAddr(handle, "glTexSubImage2D");
	pfnProxyUniform1f = glProxyGetProcAddr(handle, "glUniform1f");
	pfnProxyUniform1fv = glProxyGetProcAddr(handle, "glUniform1fv");
	pfnProxyUniform1i = glProxyGetProcAddr(handle, "glUniform1i");
	pfnProxyUniform1iv = glProxyGetProcAddr(handle, "glUniform1iv");
	pfnProxyUniform2f = glProxyGetProcAddr(handle, "glUniform2f");
	pfnProxyUniform2fv = glProxyGetProcAddr(handle, "glUniform2fv");
	pfnProxyUniform2i = glProxyGetProcAddr(handle, "glUniform2i");
	pfnProxyUniform2iv = glProxyGetProcAddr(handle, "glUniform2iv");
	pfnProxyUniform3f = glProxyGetProcAddr(handle, "glUniform3f");
	pfnProxyUniform3fv = glProxyGetProcAddr(handle, "glUniform3fv");
	pfnProxyUniform3i = glProxyGetProcAddr(handle, "glUniform3i");
	pfnProxyUniform3iv = glProxyGetProcAddr(handle, "glUniform3iv");
	pfnProxyUniform4f = glProxyGetProcAddr(handle, "glUniform4f");
	pfnProxyUniform4fv = glProxyGetProcAddr(handle, "glUniform4fv");
	pfnProxyUniform4i = glProxyGetProcAddr(handle, "glUniform4i");
	pfnProxyUniform4iv = glProxyGetProcAddr(handle, "glUniform4iv");
	pfnProxyUniformMatrix2fv = glProxyGetProcAddr(handle, "glUniformMatrix2fv");
	pfnProxyUniformMatrix3fv = glProxyGetProcAddr(handle, "glUniformMatrix3fv");
	pfnProxyUniformMatrix4fv = glProxyGetProcAddr(handle, "glUniformMatrix4fv");
	pfnProxyUseProgram = glProxyGetProcAddr(handle, "glUseProgram");
	pfnProxyValidateProgram = glProxyGetProcAddr(handle, "glValidateProgram");
	pfnProxyVertexAttrib1f = glProxyGetProcAddr(handle, "glVertexAttrib1f");
	pfnProxyVertexAttrib1fv = glProxyGetProcAddr(handle, "glVertexAttrib1fv");
	pfnProxyVertexAttrib2f = glProxyGetProcAddr(handle, "glVertexAttrib2f");
	pfnProxyVertexAttrib2fv = glProxyGetProcAddr(handle, "glVertexAttrib2fv");
	pfnProxyVertexAttrib3f = glProxyGetProcAddr(handle, "glVertexAttrib3f");
	pfnProxyVertexAttrib3fv = glProxyGetProcAddr(handle, "glVertexAttrib3fv");
	pfnProxyVertexAttrib4f = glProxyGetProcAddr(handle, "glVertexAttrib4f");
	pfnProxyVertexAttrib4fv = glProxyGetProcAddr(handle, "glVertexAttrib4fv");
	pfnProxyVertexAttribPointer = glProxyGetProcAddr(handle, "glVertexAttribPointer");
	pfnProxyViewport = glProxyGetProcAddr(handle, "glViewport");
	
	if (OPENGL_BACKEND == proxy_context.current_backend) {
		if (glxProxyInit(handle) < 0) {
			return -1;
		}
	}

	return 0;
}

extern int glProxyExtInitGLES2(void *handle);
extern int glProxyExtInitGL(void *handle);
extern int glxProxyExtInit(void *handle);

static int glProxyExtInit(void *handle)
{
	/* OpenGL extended functionality support*/
	if (NULL == handle) {
		return -1;
	}

	if (OPENGL_ES20_BACKEND == proxy_context.current_backend) {
		glProxyExtInitGLES2(handle);
	}
	else if (OPENGL_BACKEND == proxy_context.current_backend) {
		glProxyExtInitGL(handle);
		glxProxyExtInit(handle);
	}

	return 0;
}

int glProxyInit(PROXY_BACKEND_OPTION *pProxyBackend)
{
	proxy_context.gl_handle = glProxyBackendSelection(pProxyBackend);

	if (NULL == proxy_context.gl_handle) {
		printf("Error: No OpenGL or OpenGL ES2.0 driver package installed\n");
		return -1;
	}

	proxy_context.current_backend = *pProxyBackend;
	printf("%s backend is selected for rendering\n", (proxy_context.current_backend == OPENGL_BACKEND) ? "OpenGL" : "OpenGL ES2.0");

	if (OPENGL_ES20_BACKEND == proxy_context.current_backend) {
		proxy_context.egl_handle = glProxyOpenLibrary("/usr/lib/libEGL.so");
		if (NULL == proxy_context.egl_handle) {
			printf("Error: No EGL driver package installed for GLES2\n");
			return -1;
		}
	}

	glProxyEarlyResolve();

	if (glProxyCoreInit(proxy_context.gl_handle) < 0) {
		return -1;
	}

	if (glProxyExtInit(proxy_context.gl_handle) < 0) {
		return -1;
	}

	return 0;
}

int glProxyClose(void)
{
	if (NULL != proxy_context.gl_handle) {
		glProxyCloseLibrary(proxy_context.gl_handle);
	}

	if (NULL != proxy_context.egl_handle) {
		glProxyCloseLibrary(proxy_context.egl_handle);
	}

	return 0;
}
