/*
 * Copyright 2010 Linaro
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation.
 *
 * Authored by: Jammy Zhou <jammy.zhou@linaro.org>
 *
 */

#include <GL/gl_proxy.h>
#include <GL/glx_proxy.h>

PFNPROXYCHOOSEVISUALPROC            pfnProxyChooseVisual = NULL;
PFNPROXYCREATECONTEXTPROC           pfnProxyCreateContext = NULL;
PFNPROXYDESTROYCONTEXTPROC          pfnProxyDestroyContext = NULL;
PFNPROXYMAKECURRENTPROC             pfnProxyMakeCurrent = NULL;
PFNPROXYCOPYCONTEXTPROC             pfnProxyCopyContext = NULL;
PFNPROXYSWAPBUFFERSPROC             pfnProxySwapBuffers = NULL;
PFNPROXYCREATEGLXPIXMAPPROC         pfnProxyCreateGLXPixmap = NULL;
PFNPROXYDESTROYGLXPIXMAPPROC        pfnProxyDestroyGLXPixmap = NULL;
PFNPROXYQUERYEXTENSIONPROC          pfnProxyQueryExtension = NULL;
PFNPROXYQUERYVERSIONPROC            pfnProxyQueryVersion = NULL;
PFNPROXYISDIRECTPROC                pfnProxyIsDirect = NULL;
PFNPROXYGETCONFIGPROC               pfnProxyGetConfig = NULL;
PFNPROXYGETCURRENTCONTEXTPROC       pfnProxyGetCurrentContext = NULL;
PFNPROXYGETCURRENTDRAWABLEPROC      pfnProxyGetCurrentDrawable = NULL;
PFNPROXYWAITGLPROC                  pfnProxyWaitGL = NULL;
PFNPROXYWAITXPROC                   pfnProxyWaitX = NULL;
PFNPROXYUSEXFONTPROC                pfnProxyUseXFont = NULL;
PFNPROXYQUERYEXTENSIONSTRINGPROC    pfnProxyQueryExtensionString = NULL;
PFNPROXYQUERYSERVERSTRINGPROC       pfnProxyQueryServerString = NULL;
PFNPROXYGETCLIENTSTRINGPROC         pfnProxyGetClientString = NULL;
PFNPROXYGETCURRENTDISPLAYPROC       pfnProxyGetCurrentDisplay = NULL;
PFNPROXYCHOOSEFBCONFIGPROC          pfnProxyChooseFBConfig = NULL;
PFNPROXYGETFBCONFIGATTRIBPROC       pfnProxyGetFBConfigAttrib = NULL;
PFNPROXYGETFBCONFIGSPROC            pfnProxyGetFBConfigs = NULL;
PFNPROXYGETVISUALFROMFBCONFIGPROC   pfnProxyGetVisualFromFBConfig = NULL;
PFNPROXYCREATEWINDOWPROC            pfnProxyCreateWindow = NULL;
PFNPROXYDESTROYWINDOWPROC           pfnProxyDestroyWindow = NULL;
PFNPROXYCREATEPIXMAPPROC            pfnProxyCreatePixmap = NULL;
PFNPROXYDESTROYPIXMAPPROC           pfnProxyDestroyPixmap = NULL;
PFNPROXYCREATEPBUFFERPROC           pfnProxyCreatePBuffer = NULL;
PFNPROXYDESTROYPBUFFERPROC          pfnProxyDestroyPBuffer = NULL;
PFNPROXYQUERYDRAWABLEPROC           pfnProxyQueryDrawable = NULL;
PFNPROXYCREATENEWCONTEXTPROC        pfnProxyCreateNewContext = NULL;
PFNPROXYMAKECONTEXTCURRENTPROC      pfnProxyMakeContextCurrent = NULL;
PFNPROXYGETCURRENTREADDRAWABLEPROC  pfnProxyGetCurrentReadDrawable = NULL;
PFNPROXYQUERYCONTEXTPROC            pfnProxyQueryContext = NULL;
PFNPROXYSELECTEVENTPROC             pfnProxySelectEvent = NULL;
PFNPROXYGETSELECTEDEVENTPROC        pfnProxyGetSelectedEvent = NULL;

extern void *glProxyGetProcAddr(void *handle, const char *name);

int glxProxyInit(void *handle)
{
	if (NULL == handle) {
                return -1;
        }

        pfnProxyChooseVisual = glProxyGetProcAddr(handle, "glXChooseVisual");
        pfnProxyCreateContext = glProxyGetProcAddr(handle, "glXCreateContext");
        pfnProxyDestroyContext = glProxyGetProcAddr(handle, "glXDestroyContext");
        pfnProxyMakeCurrent = glProxyGetProcAddr(handle, "glXMakeCurrent");
        pfnProxyCopyContext = glProxyGetProcAddr(handle, "glXCopyContext");
        pfnProxySwapBuffers = glProxyGetProcAddr(handle, "glXSwapBuffers");
        pfnProxyCreateGLXPixmap = glProxyGetProcAddr(handle, "glXCreateGLXPixmap");
        pfnProxyDestroyGLXPixmap = glProxyGetProcAddr(handle, "glXDestroyGLXPixmap");
        pfnProxyQueryExtension = glProxyGetProcAddr(handle, "glXQueryExtension");
        pfnProxyQueryVersion = glProxyGetProcAddr(handle, "glXQueryVersion");
        pfnProxyIsDirect = glProxyGetProcAddr(handle, "glXIsDirect");
        pfnProxyGetConfig = glProxyGetProcAddr(handle, "glXGetConfig");
        pfnProxyGetCurrentContext = glProxyGetProcAddr(handle, "glXGetCurrentContext");
        pfnProxyGetCurrentDrawable = glProxyGetProcAddr(handle, "glXGetCurrentDrawable");
        pfnProxyWaitGL = glProxyGetProcAddr(handle, "glXWaitGL");
        pfnProxyWaitX = glProxyGetProcAddr(handle, "glXWaitX");     
        pfnProxyUseXFont = glProxyGetProcAddr(handle, "glXUseXFont");
        pfnProxyQueryExtensionString = glProxyGetProcAddr(handle, "glXQueryExtensionString");
        pfnProxyQueryServerString = glProxyGetProcAddr(handle, "glXQueryServerString");
        pfnProxyGetClientString = glProxyGetProcAddr(handle, "glXGetClientString");
        pfnProxyGetCurrentDisplay = glProxyGetProcAddr(handle, "glXGetCurrentDisplay");
        pfnProxyChooseFBConfig = glProxyGetProcAddr(handle, "glXChooseFBConfig");
        pfnProxyGetFBConfigAttrib = glProxyGetProcAddr(handle, "glXGetFBConfigAttrib");
        pfnProxyGetFBConfigs = glProxyGetProcAddr(handle, "glXGetFBConfigs");
        pfnProxyGetVisualFromFBConfig = glProxyGetProcAddr(handle, "glXGetVisualFromFBConfig");
        pfnProxyCreateWindow = glProxyGetProcAddr(handle, "glXCreateWindow");
        pfnProxyDestroyWindow = glProxyGetProcAddr(handle, "glXDestroyWindow");
        pfnProxyCreatePixmap = glProxyGetProcAddr(handle, "glXCreatePixmap");
        pfnProxyDestroyPixmap = glProxyGetProcAddr(handle, "glXDestroyPixmap");
        pfnProxyCreatePBuffer = glProxyGetProcAddr(handle, "glXCreatePBuffer");
        pfnProxyDestroyPBuffer = glProxyGetProcAddr(handle, "glXDestroyPBuffer");
        pfnProxyQueryDrawable = glProxyGetProcAddr(handle, "glXQueryDrawable");
        pfnProxyCreateNewContext = glProxyGetProcAddr(handle, "glXCreateNewContext");
        pfnProxyMakeContextCurrent = glProxyGetProcAddr(handle, "glXMakeContextCurrent");
        pfnProxyGetCurrentReadDrawable = glProxyGetProcAddr(handle, "glXGetCurrentReadDrawable");
        pfnProxyQueryContext = glProxyGetProcAddr(handle, "glXQueryContext");
        pfnProxySelectEvent = glProxyGetProcAddr(handle, "glXSelectEvent");
        pfnProxyGetSelectedEvent = glProxyGetProcAddr(handle, "glXGetSelectecEvent");

	return 0;
}

int glxProxyExtInit(void *handle)
{
	/* TODO */
	return 0;
}	
